from importlib import import_module
from types import ModuleType


def import_pkg(base: str, sub: str) -> ModuleType:
    """
    Import the package located at base.str
    :param base:
    :param sub:
    :return:
    """
    pkg_path = f"{base}.{sub}"
    try:
        pkg = import_module(pkg_path)
    except ModuleNotFoundError as _e:
        raise Exception(
            f"Cannot find '{pkg_path}' on the current path. Check that the module '{base}' is on "
            f"'sys.path', and that it contains the sub-module '{sub}'."
        )
    return pkg



import numpy as np
from typing import Tuple

from torch.utils.data import DataLoader, Dataset
from hijax.setup.utils import import_pkg


def worker_init_fn(worker_id):
    """
    Ensures that shuffle order and data augmentation is varied
    between the threads, when num_workers > 1.
    Note: possibly fixed in PyTorch 1.9
    :param worker_id:
    :return:
    """
    np.random.seed(np.random.get_state()[1][0] + worker_id)


def setup_loaders(dataset_class: str, data_opts: dict, loader_opts: dict, module: str) -> Tuple[DataLoader, DataLoader]:
    """
    Create PyTorch data loaders for training and validation.
    :param dataset_class: name of dataset class. any class exported in <module>/data/__init__.py
    :param data_opts: `dataset` config sub-dictionary
    :param loader_opts: `loader` config sub-dictionary
    :param module: string name of the module to import `datasets` from
    :return:
    """
    # load available datasets from specified module
    datasets = import_pkg(module, "datasets")

    # get dataset options
    both = data_opts.get("both", {})
    train = data_opts.get("train", {})
    test = data_opts.get("test", {})

    # initialise datasets
    dataset_class = getattr(datasets, dataset_class)
    train_dataset: Dataset = dataset_class(**{**both, **train})
    test_dataset: Dataset = dataset_class(**{**both, **test})

    # loader settings
    both = loader_opts.get("both", {})
    train = loader_opts.get("train", {})
    test = loader_opts.get("test", {})

    # account for custom `collate_fn`
    collate_fn = None
    if hasattr(train_dataset, "get_collate_fn"):
        collate_fn = train_dataset.get_collate_fn()

    # account for custom batch sampler
    train_loader_opts = {**both, **train}
    test_loader_opts = {**both, **test}
    if hasattr(train_dataset, "get_batch_sampler"):
        train_loader_opts["batch_sampler"] = train_dataset.get_batch_sampler()
        test_loader_opts["batch_sampler"] = test_dataset.get_batch_sampler()

        # remove mutually exclusive args
        for opts in (train_loader_opts, test_loader_opts):
            opts.pop("batch_size", None)
            opts.pop("shuffle", None)
            opts.pop("sampler", None)
            opts.pop("drop_last", None)

    # initialise loaders
    train_loader = DataLoader(
        dataset=train_dataset, collate_fn=collate_fn, worker_init_fn=worker_init_fn, **train_loader_opts
    )
    test_loader = DataLoader(
        dataset=test_dataset, collate_fn=collate_fn, worker_init_fn=worker_init_fn, **test_loader_opts
    )

    return train_loader, test_loader



import os
import wandb

from omegaconf import DictConfig
from typing import Optional, Union

import torch
import haiku as hk

from hijax.setup.setup_config import setup_config
from hijax.setup.utils import import_pkg


def setup_worker(
    name: str,
    module: str,
    cfg: Optional[Union[DictConfig, dict]] = None,
    exp_dir: Optional[str] = None,
    include_wandb: bool = True,
    overwrite: bool = False,
    reset_metrics: bool = False,
    checkpoint_id: str = "best",
):
    """
    Create or load an experiment.
    :param name: unique experiment name for saving/resuming
    :param module: a python module containing `workers`, `models` and `datasets`
    :param cfg: experiment config. can be `None` for existing experiment.
        otherwise a `dict` or `DictConfig` containing the configuration.
    :param exp_dir: directory to store experiment runs
    :param include_wandb: include an instance of wandb ? (required for training)
    :param overwrite: overwrite existing experiment
    :param reset_metrics: if `True` the `lowest_loss` and `summary_stats` are reset
    :param checkpoint_id: for inference / resuming, which checkpoint should we load?
    :return:
    """
    # load the available workers and models
    workers = import_pkg(module, "workers")
    models = import_pkg(module, "models")

    # get experiment directory
    if exp_dir is None:
        exp_dir = os.environ.get("EXPERIMENT_DIR", False)
        if not exp_dir:
            raise Exception(
                "No experiment directory defined. Set environment variable `EXPERIMENT_DIR` or "
                "pass as kwarg `setup_worker(..., exp_dir=?)"
            )

    # setup the directory
    cfg = setup_config(name, cfg, exp_dir, overwrite)

    # initialise model
    model_class = getattr(models, cfg["model"].pop("class"))
    model: Union[torch.nn.Module, hk.Module] = model_class(**cfg["model"])

    # setup visualisation
    run = None
    if include_wandb and "wandb" in cfg:
        cfg_ = cfg
        if type(cfg) == DictConfig:
            cfg_ = cfg.to_dict()
        run = wandb.init(name=name, config=cfg_, id=cfg["run_id"], resume="allow", **cfg["wandb"])

        # if using PyTorch, WandB can track the gradients automatically like this...
        if type(model) == torch.nn.Module:
            run.watch(model)

    # initialise the worker
    run_dir = "{}/{}".format(exp_dir, name)
    worker_class = getattr(workers, cfg["worker"].pop("class"))
    worker = worker_class(
        name, model, run_dir, run, checkpoint_id=checkpoint_id, reset_metrics=reset_metrics, **cfg["worker"]
    )

    return worker, cfg



import torch
import torch.nn as nn
import os
import numpy as np
from abc import ABC, abstractmethod
from typing import Any, Optional
from pathlib import Path

from wandb.wandb_run import Run

from torch.utils.data import DataLoader


class AbstractWorker(ABC):
    def __init__(
        self,
        exp_name: str,
        run_dir: str,
        wandb: Optional[Run] = None,
        epoch_save_freq: int = 50,
        log_interval: int = 50,
        *_args,
        **kwargs,
    ):
        """
        All workers should inherit from this class. It will:
        1. provide common utilities for saving, loading and device management.
        2. enforce a common interface between workers. i.e) must implement `.train()` and `evaluate.()`
        3. provide a common interface for logging to WandB.

        :param exp_name: name of experiment
        :param run_dir: directory to save configurations, weights, artifacts for this experiment.
        :param wandb: the session that is returned from wandb.init(...).
        :param epoch_save_freq: how often to save intermediate checkpoints
        :param log_interval: how many gradient updates before logging to wandb
        :param args:
        :param kwargs:
        """

        self.reset_metrics = kwargs.get("reset_metrics", False)
        self.exp_name = exp_name
        self.wandb = wandb
        self.run_dir = run_dir
        self.upload_checkpoints = kwargs.get("upload_checkpoints", False)

        # keep a list of all stateful objects, to include in the experiment checkpoints
        # includes the model by default
        # see: `register_state()
        self.stateful_objects = {}

        # track loss and summary stats
        self.lowest_loss = float("inf")
        self.summary_stats = {}

        # epoch related stuff
        self.epoch_counter = 0
        self.epoch_save_freq = epoch_save_freq
        self.log_interval = log_interval

        # cache values until its time to plot them
        self._counters = {"train": 0, "test": 0}
        self._metric_cache = {"train": {}, "test": {}}

    @abstractmethod
    def train(self, loader):
        pass

    @abstractmethod
    def evaluate(self, loader):
        pass

    def run(self, train_loader: DataLoader, test_loader: DataLoader, nb_epoch: int):
        """
        Run an experiment for the specified number of epoch.
        :param train_loader:
        :param test_loader:
        :param nb_epoch:
        :return:
        """
        for epoch in range(nb_epoch):
            # reset numpy random seed
            # TODO: check how we are handling random state for Haiku
            np.random.seed(0)

            # train and test sets
            self.train(train_loader)
            self.epoch_counter += 1
            with torch.no_grad():
                loss_score, summary_stats = self.evaluate(test_loader)

            # track individual metrics at an epoch level
            assert type(summary_stats) == dict, "`worker.evaluate()` expects return type : (float, Dict)"
            for k, v in summary_stats.items():
                self.summary_stats[k] = v

            # save `best`
            if loss_score < self.lowest_loss:
                print("New lowest test loss {}".format(loss_score))
                self.save(checkpoint_id="best")
                self.lowest_loss = loss_score
                if self.wandb is not None:
                    self.wandb.summary["lowest_loss"] = loss_score
                    for k, v in self.summary_stats.items():
                        self.wandb.summary[f"{k}: Lowest Test Set (Avg)"] = v

            # save every `x` epoch
            if self.epoch_counter % self.epoch_save_freq == 0:
                self.save(checkpoint_id="{}".format(self.epoch_counter))

            # overwrite latest weights
            self.save(checkpoint_id="latest")

    def cuda(self, device_id: int = 0):
        """
        Move the model and all optimisers to the GPU.
        :return:
        """
        assert torch.cuda.is_available(), "CUDA support not found!"
        for key, obj in self.stateful_objects.items():
            if isinstance(obj, AbstractWorker):
                continue
            if hasattr(obj, "cuda"):
                self.stateful_objects[key].cuda(device_id)

    def register_state(self, obj: Any, name: str):
        """
        Register an object to be included in the experiment checkpoint.
        Any object implementing `.load_state_dict()` and `.state_dict()` is valid. For example:
            - PyTorch optimiser, LR Scheduler or Model
            - Any custom class implementing those methods (e.g. a specific worker with extra state)

        :param obj: obj
        :param name: unique name for what is being registered
        """
        assert name not in self.stateful_objects, "Duplicate key in state list for '{}'".format(name)
        assert hasattr(obj, "load_state_dict"), "Object must implement `load_state_dict` to be included in checkpoint."
        assert hasattr(obj, "state_dict"), "Object must implement `state_dict` to be included in checkpoint."

        self.stateful_objects[name] = obj

    def save(self, checkpoint_id: str = "latest"):
        """
        Save state of all tracked modules.
        :param checkpoint_id:
        :return:
        """
        state_dict = {
            "lowest_loss": self.lowest_loss,
            "summary_stats": self.summary_stats,
            "epoch_counter": self.epoch_counter,
        }
        for key, obj in self.stateful_objects.items():
            state_dict[key] = obj.state_dict()

        checkpoint_path = "{}/checkpoint_{}.pt".format(self.run_dir, checkpoint_id)
        print("Saving checkpoint {}".format(checkpoint_path))
        torch.save(state_dict, checkpoint_path)

        # upload and overwrite the checkpoints to wandb if `upload_checkpoints` enabled in worker config
        if self.wandb is not None and self.upload_checkpoints and checkpoint_id in ["latest", "best"]:
            self.wandb.save(glob_str=checkpoint_path, base_path=str(Path(self.run_dir).parent), policy="live")

    def load(self, checkpoint_id: str = "best", strict: bool = True, with_optim: bool = True):
        """
        Load state from existing checkpoint.
        :param checkpoint_id:
        :param strict: whether to use strict loading for the model weights (see PyTorch nn.Module `load_state_dict)
        :param with_optim:
        :return:
        """

        checkpoint_path = "{}/checkpoint_{}.pt".format(self.run_dir, checkpoint_id)

        if not os.path.exists(checkpoint_path):
            print("Checkpoint not found {}".format(checkpoint_path))
            return

        print("Loading checkpoint {}".format(checkpoint_path))

        state_dict = torch.load(checkpoint_path)
        for key, state in state_dict.items():
            # skip optimiser
            if not with_optim and key == "optim":
                continue
            # skip over important top-level state (always track this!)
            if key in {"lowest_loss", "summary_stats", "epoch_counter"}:
                continue
            # handle remaining objects (model, worker, optimiser, any other user defined stuff)
            if key not in self.stateful_objects:
                print(
                    f"Warning! {key} was found in the checkpoint but not in the stateful objects for the current"
                    f"run!"
                )
            else:
                obj = self.stateful_objects[key]
                if isinstance(obj, nn.Module):
                    obj.load_state_dict(state, strict)
                else:
                    obj.load_state_dict(state)

        if not self.reset_metrics:
            self.summary_stats = state_dict.get("summary_stats", {})
            self.lowest_loss = state_dict.get("lowest_loss", float("inf"))
        self.epoch_counter = state_dict.get("epoch_counter", 0)

    @staticmethod
    def unwrap_value(v):
        if torch.is_tensor(v):
            return v.item()
        elif type(v) == np.ndarray:
            return v.item()
        else:
            return v

    def _plot_loss(self, metrics: dict, train=True):
        """
        Plot metrics to weights and biases.

        Keeps a moving average of values and pushes them every `log_interval`.

        :param metrics: dictionary of things to track
        :param train: ?
        """
        subset = "train" if train else "test"

        for k, v in metrics.items():
            full_key = "{} {}".format(k, subset)
            if full_key not in self._metric_cache[subset]:
                self._metric_cache[subset][full_key] = []
            v_raw = AbstractWorker.unwrap_value(v)
            self._metric_cache[subset][full_key].append(v_raw)

        # increment the counter
        self._counters[subset] += 1

        # empty cache and plot!
        if self._counters[subset] % self.log_interval == 0:
            avg = {}
            for k, v in self._metric_cache[subset].items():
                avg[k] = np.mean(v)
                self._metric_cache[subset][k] = []

            self.wandb.log(avg)
            self._counters[subset] = 0



import os
from typing import Optional, Union

import wandb
from omegaconf import DictConfig, OmegaConf


def setup_config(
    name: str,
    cfg: Optional[Union[DictConfig, dict]] = None,
    exp_dir: Optional[str] = None,
    overwrite: bool = False,
    verbose: bool = True,
) -> DictConfig:
    """
    Create an experiment with the specific name and settings
    OR create a new experiment if it doesn't exist
    OR overwrite the configuration of an existing experiment

    :param name: unique experiment name for saving/resuming
    :param cfg: experiment config. can be `None` for existing experiment.
        otherwise a `dict` or `DictConfig` containing the configuration.
    :param exp_dir: directory to store experiment runs
    :param overwrite: overwrite existing experiment
    :param verbose: print out the loaded config
    :return:
    """
    # if a regular `dict` type is passed in, convert to `DictConfig` to make use of the YAML
    # serialization methods
    if type(cfg) == dict:
        cfg = DictConfig(cfg)

    # create the experiment directory if it doesn't exist
    if not os.path.exists(exp_dir):
        os.mkdir(exp_dir)

    # create the directory for this specific run
    run_dir = "{}/{}".format(exp_dir, name)
    if not os.path.exists(run_dir):
        os.mkdir(run_dir)

    # store `config.yaml` alongside the weights
    config_path = "{}/config.yaml".format(run_dir)
    if not os.path.exists(config_path):
        assert cfg is not None, "No config. supplied and no existing experiment found at {}".format(config_path)
        print("Creating new experiment.")

        # generate a unique id for this run
        run_id = wandb.util.generate_id()
        OmegaConf.set_struct(cfg, False)
        cfg["run_id"] = run_id

        OmegaConf.save(cfg, config_path)
    # Note: when overwriting parameters we will preserve the same `run_id`. This allows to resume
    # training with tweaked parameters (e.g lower learning rate)
    else:
        existing_cfg = OmegaConf.load(config_path)
        run_id = existing_cfg["run_id"]

        if not overwrite:
            print("Loading existing experiment at {}".format(config_path))
            print("If you would like to overwrite, please set overwrite=True.")
            cfg = existing_cfg
        else:
            print("Overwriting existing configuration at {}".format(config_path))
            OmegaConf.set_struct(cfg, False)
            cfg["run_id"] = run_id
            OmegaConf.save(cfg, config_path)

    # print the configuration
    if verbose:
        print(OmegaConf.to_yaml(cfg))

    return cfg



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hijax', 'hijax.setup']

package_data = \
{'': ['*']}

install_requires = \
['black>=21.10b0,<22.0',
 'dm-haiku>=0.0.5,<0.0.6',
 'hydra-core>=1.1.1,<2.0.0',
 'jax>=0.2.24,<0.3.0',
 'mypy>=0.910,<0.911',
 'numpy>=1.21.4,<2.0.0',
 'optax>=0.0.9,<0.0.10',
 'torch>=1.10.0,<2.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'wandb>=0.12.6,<0.13.0']

setup_kwargs = {
    'name': 'hijax',
    'version': '0.1.0',
    'description': 'An experiment framework for Haiku and Jax',
    'long_description': None,
    'author': 'Angus Turner',
    'author_email': 'angusturner27@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
